//////////////////////////////////////////////////////////////////////////////
//
//  (C) Copyright Ladislav Sopko, 3D Informatica s.r.l 2014
//  
//////////////////////////////////////////////////////////////////////////////
#if defined(_MSC_VER)&&(_MSC_VER>=1200)
#pragma once
#endif


#ifndef XW_IPC_FIXABLE_ROBUST_MUTEX_HPP
#define XW_IPC_FIXABLE_ROBUST_MUTEX_HPP


#include <xw-incl/ipc/exception.hpp>
#include <xw-incl/ipc/named_robust_mutex.hpp>
#include <xw-incl/ipc/robust_mutex_fixer.hpp>
#include <boost/interprocess/creation_tags.hpp>


#ifdef strRev
#	undef strRev
#endif
#define strRev "$Revision: 1.4 $"

namespace bipc = boost::interprocess;

namespace xw{ namespace ipc{

	template<typename mutexT>
	class scoped_custom_fix_lock : private boost::noncopyable{
		bool _locked;
		mutexT& _mutex;
	public:
		scoped_custom_fix_lock(mutexT&, const mutex_fixer_ptr_t&);
		~scoped_custom_fix_lock();
	};

	template<typename mutexT>
	scoped_custom_fix_lock<mutexT>::scoped_custom_fix_lock(mutexT& m, const mutex_fixer_ptr_t& f)
		: _locked(false), _mutex(m)
	{
		_mutex.lock(f);
		_locked = true;
	}

	template<typename mutexT>
	scoped_custom_fix_lock<mutexT>::~scoped_custom_fix_lock()
	{
		if(_locked){
			try{
				_mutex.unlock();
			}catch(...){};			
		}
	}

	template<typename mutexT>
	class fixable_robust_mutex {

		typedef typename boost::shared_ptr<mutexT> mutex_ptr_t;

		mutex_ptr_t _mutex;
		mutex_fixer_ptr_t _fixer;
		fix_it_trait<mutexT> _fix_it_trait;

		//Non-copyable
		
		fixable_robust_mutex();
		fixable_robust_mutex(const fixable_robust_mutex &);
		fixable_robust_mutex &operator=(const fixable_robust_mutex &);

		void handle_fixing(const mutex_fixer_ptr_t&);
		void handle_fixing();
		
	public:
		
		fixable_robust_mutex(
				bipc::create_only_t create_only, 
				const char *name, 
				const mutex_fixer_ptr_t& fixer = IPC_MAKE_FIXER(ex_ipc_fixer_generic_exception_fixer_t), // mutex_fixer_ptr_t(new ex_ipc_fixer_generic_exception_fixer_t()), 
				const bipc::permissions &perm = bipc::permissions()
		);

		fixable_robust_mutex(
				bipc::open_or_create_t open_or_create, 
				const char *name, 
				const mutex_fixer_ptr_t& fixer = IPC_MAKE_FIXER(ex_ipc_fixer_generic_exception_fixer_t), // mutex_fixer_ptr_t(new ex_ipc_fixer_generic_exception_fixer_t()), 
				const bipc::permissions &perm = bipc::permissions()
		);

		fixable_robust_mutex(
				bipc::open_only_t open_only, 
				const char *name, 
				const mutex_fixer_ptr_t& fixer = IPC_MAKE_FIXER(ex_ipc_fixer_generic_exception_fixer_t) // mutex_fixer_ptr_t(new ex_ipc_fixer_generic_exception_fixer_t())
		);

		~fixable_robust_mutex();

		void lock(const mutex_fixer_ptr_t&);
		void lock();
		void unlock();
		bool try_lock();
		bool timed_lock(const boost::posix_time::ptime&);

		//!Erases a named mutex from the system.
		//!Returns false on error. Never throws.
		static bool remove(const char *);

		const char* name() const;
		
	};

	/************************************************************************/
	/*  IPC types                                                           */
	/************************************************************************/

	typedef fixable_robust_mutex<robust_ipc_named_mutex> fixable_ipc_robust_named_mutex;
	typedef fixable_robust_mutex<robust_ipc_named_recursive_mutex> fixable_ipc_robust_named_recursive_mutex;

	
	/************************************************************************/
	/*  IMPL                                                                */
	/************************************************************************/

	template<typename mutexT>
	inline fixable_robust_mutex<mutexT>::fixable_robust_mutex(
		bipc::create_only_t, 
		const char *name, 
		const mutex_fixer_ptr_t& fixer, 
		const bipc::permissions &perm = bipc::permissions()) 
		: _mutex(new mutexT(bipc::create_only, name, perm)), _fixer(fixer) { 
	}

	template<typename mutexT>
	inline fixable_robust_mutex<mutexT>::fixable_robust_mutex(
		bipc::open_or_create_t, 
		const char *name, 
		const mutex_fixer_ptr_t& fixer, 
		const bipc::permissions &perm = bipc::permissions()) 
		: _mutex(new mutexT(bipc::open_or_create, name, perm)), _fixer(fixer) { 
	}

	template<typename mutexT>
	inline fixable_robust_mutex<mutexT>::fixable_robust_mutex(
		bipc::open_only_t, 
		const char *name, 
		const mutex_fixer_ptr_t& fixer) 
		: _mutex(new mutexT(bipc::open_only, name)), _fixer(fixer) { 
	}

	template<typename mutexT>
	inline fixable_robust_mutex<mutexT>::~fixable_robust_mutex(){ 
	}

	template<typename mutexT>
	inline bool fixable_robust_mutex<mutexT>::remove(const char *name)
	{  return bipc::shared_memory_object::remove(name); }

	template<typename mutexT>
	inline const char* fixable_robust_mutex<mutexT>::name() const{
		//it will unlock just correct mutex
		return _mutex->name();
	}

	template<typename mutexT>
	inline void fixable_robust_mutex<mutexT>::unlock(){
		//it will unlock just correct mutex
		if(_mutex->get_state() == robust_mutex_state::correct_state)
			_mutex->unlock();
	}

	template<typename mutexT>
	inline void fixable_robust_mutex<mutexT>::lock(){
		_mutex->lock();
		handle_fixing();
	}

	template<typename mutexT>
	inline void fixable_robust_mutex<mutexT>::lock(const mutex_fixer_ptr_t& fixer){
		_mutex->lock();
		handle_fixing(fixer);
	}

	template<typename mutexT>
	inline bool fixable_robust_mutex<mutexT>::try_lock(){
		bool ret = _mutex->try_lock();
		if(ret) handle_fixing();	
		return ret;
	}

	template<typename mutexT>
	inline bool fixable_robust_mutex<mutexT>::timed_lock(const boost::posix_time::ptime & t){
		bool ret = _mutex->timed_lock(t);
		if(ret) handle_fixing();	
		return ret;
	}

	template<typename mutexT>
	inline void fixable_robust_mutex<mutexT>::handle_fixing(){
		handle_fixing(_fixer);
	}

	template<typename mutexT>
	inline void fixable_robust_mutex<mutexT>::handle_fixing(const mutex_fixer_ptr_t& f){
		//check eventual fixing state
		if(_mutex->get_state() == robust_mutex_state::fixing_state){
			if(f->fix()) _fix_it_trait.fix_it(*_mutex);
		}
	}

}}  //xw::ipc

#endif
