//////////////////////////////////////////////////////////////////////////////
//
// (C) Copyright Ion Gaztanaga 2005-2012. Distributed under the Boost
// Software License, Version 1.0. (See accompanying file
// LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//
// See http://www.boost.org/libs/interprocess for documentation.
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
//  (C) Copyright Ladislav Sopko, 3D Informatica s.r.l 2014
//  
//////////////////////////////////////////////////////////////////////////////
#if (defined _MSC_VER) && (_MSC_VER >= 1200)
#  pragma once
#endif

#ifndef XW_IPC_ROBUST_SHM_NAMED_MUTEX_HPP
#define XW_IPC_ROBUST_SHM_NAMED_MUTEX_HPP

#include <boost/interprocess/creation_tags.hpp>
#include <boost/interprocess/permissions.hpp>
#include <boost/interprocess/detail/posix_time_types_wrk.hpp>

#include <boost/interprocess/shared_memory_object.hpp>
#include <boost/interprocess/detail/managed_open_or_create_impl.hpp>
#include <boost/interprocess/sync/shm/named_creation_functor.hpp>
#include <xw-incl/ipc/exception.hpp>
#include <xw-incl/ipc/robust_recursive_mutex.hpp>

#ifdef strRev
#	undef strRev
#endif
#define strRev "$Revision: 1.4 $"

//!\file
//!Describes a named mutex class for inter-process synchronization

namespace bipc = boost::interprocess;

namespace xw { namespace ipc {

	//!A mutex with a global name, so it can be found from different
	//!processes. This mutex can't be placed in shared memory, and
	//!each process should have it's own named mutex.
	template<typename mutexT>
	class robust_shm_named_mutex
	{
		typedef typename mutexT internal_mutex_type;
		typedef typename robust_shm_named_mutex<mutexT> self_type;

		friend class fix_it_trait<self_type>;

		//Non-copyable	
		
		robust_shm_named_mutex();
		robust_shm_named_mutex(const robust_shm_named_mutex &);
		robust_shm_named_mutex &operator=(const robust_shm_named_mutex &);
		
	public:
		//!Creates a global interprocess_mutex with a name.
		//!Throws interprocess_exception on error.
		robust_shm_named_mutex(bipc::create_only_t create_only, const char *name, const bipc::permissions &perm = bipc::permissions());

		//!Opens or creates a global mutex with a name.
		//!If the mutex is created, this call is equivalent to
		//!shm_named_mutex(create_only_t, ... )
		//!If the mutex is already created, this call is equivalent
		//!shm_named_mutex(open_only_t, ... )
		//!Does not throw
		robust_shm_named_mutex(bipc::open_or_create_t open_or_create, const char *name, const bipc::permissions &perm = bipc::permissions());

		//!Opens a global mutex with a name if that mutex is previously
		//!created. If it is not previously created this function throws
		//!interprocess_exception.
		robust_shm_named_mutex(bipc::open_only_t open_only, const char *name);

		//!Destroys *this and indicates that the calling process is finished using
		//!the resource. The destructor function will deallocate
		//!any system resources allocated by the system for use by this process for
		//!this resource. The resource can still be opened again calling
		//!the open constructor overload. To erase the resource from the system
		//!use remove().
		~robust_shm_named_mutex();

		//!Unlocks a previously locked
		//!interprocess_mutex.
		void unlock();

		//!Locks interprocess_mutex, sleeps when interprocess_mutex is already locked.
		//!Throws interprocess_exception if a severe error is found
		void lock();

		//!Tries to lock the interprocess_mutex, returns false when interprocess_mutex
		//!is already locked, returns true when success.
		//!Throws interprocess_exception if a severe error is found
		bool try_lock();

		//!Tries to lock the interprocess_mutex until time abs_time,
		//!Returns false when timeout expires, returns true when locks.
		//!Throws interprocess_exception if a severe error is found
		bool timed_lock(const boost::posix_time::ptime &abs_time);

		//!Erases a named mutex from the system.
		//!Returns false on error. Never throws.
		static bool remove(const char *name);

		boost::uint32_t get_state();
		
		const char* name() const;
		
		internal_mutex_type &internal_mutex()
		{  return *static_cast<internal_mutex_type*>(m_shmem.get_user_address()); }

		const internal_mutex_type &internal_mutex() const
		{  return *static_cast<internal_mutex_type*>(m_shmem.get_user_address()); }

	private:
		void fix_it(); //who need call must have trait for it!!!

		typedef bipc::ipcdetail::managed_open_or_create_impl<bipc::shared_memory_object, 0, true, false> open_create_impl_t;
		open_create_impl_t m_shmem;
		typedef bipc::ipcdetail::named_creation_functor<internal_mutex_type> construct_func_t;
		set_name_trait<mutexT> _nameTrait;
		fix_it_trait<mutexT> _fixItTrait; //trough this we can go on private function in wrapped mutex
	};

	/************************************************************************/
	/*  IPC types                                                           */
	/************************************************************************/

	typedef robust_shm_named_mutex<robust_ipc_mutex> robust_ipc_named_mutex;
	typedef boost::shared_ptr<robust_ipc_named_mutex> robust_ipc_named_mutex_ptr_t;

	typedef robust_shm_named_mutex<robust_ipc_recursive_mutex> robust_ipc_named_recursive_mutex;
	typedef boost::shared_ptr<robust_ipc_named_mutex> robust_ipc_named_recursive_mutex_ptr_t;
	
	/************************************************************************/
	/*  IMPL                                                                */
	/************************************************************************/

	template<typename mutexT>
	inline robust_shm_named_mutex<mutexT>::~robust_shm_named_mutex()
	{}

	template<typename mutexT>
	inline robust_shm_named_mutex<mutexT>::robust_shm_named_mutex(bipc::create_only_t, const char *name, const bipc::permissions &perm)
		:  m_shmem  (bipc::create_only
		,name
		,sizeof(internal_mutex_type) +
		open_create_impl_t::ManagedOpenOrCreateUserOffset
		,bipc::read_write
		,0
		,construct_func_t(bipc::ipcdetail::DoCreate)
		,perm)
	{
		_nameTrait.name(this->internal_mutex(), name);
	}

	template<typename mutexT>
	inline robust_shm_named_mutex<mutexT>::robust_shm_named_mutex(bipc::open_or_create_t, const char *name, const bipc::permissions &perm)
		:  m_shmem  (bipc::open_or_create
		,name
		,sizeof(internal_mutex_type) +
		open_create_impl_t::ManagedOpenOrCreateUserOffset
		,bipc::read_write
		,0
		,construct_func_t(bipc::ipcdetail::DoOpenOrCreate)
		,perm)
	{
		_nameTrait.name(this->internal_mutex(), name);	
	}

	template<typename mutexT>
	inline robust_shm_named_mutex<mutexT>::robust_shm_named_mutex(bipc::open_only_t, const char *name)
		:  m_shmem  (bipc::open_only
		,name
		,bipc::read_write
		,0
		,construct_func_t(bipc::ipcdetail::DoOpen))
	{
		_nameTrait.name(this->internal_mutex(), name);
	}

	template<typename mutexT>
	inline void robust_shm_named_mutex<mutexT>::lock()
	{  this->internal_mutex().lock();  }

	template<typename mutexT>
	inline const char* robust_shm_named_mutex<mutexT>::name() const
	{  return this->internal_mutex().name();  }

	template<typename mutexT>
	inline void robust_shm_named_mutex<mutexT>::unlock()
	{  this->internal_mutex().unlock();  }

	template<typename mutexT>
	inline bool robust_shm_named_mutex<mutexT>::try_lock()
	{  return this->internal_mutex().try_lock();  }

	template<typename mutexT>
	inline bool robust_shm_named_mutex<mutexT>::timed_lock(const boost::posix_time::ptime &abs_time)
	{
		if(abs_time == boost::posix_time::pos_infin){
			this->lock();
			return true;
		}
		return this->internal_mutex().timed_lock(abs_time);
	}

	template<typename mutexT>
	inline bool robust_shm_named_mutex<mutexT>::remove(const char *name)
	{  return bipc::shared_memory_object::remove(name); }

	template<typename mutexT>
	inline void robust_shm_named_mutex<mutexT>::fix_it()
	{ 
		//use trait to fix mutex, fix_it is private in mutex !!!, trait is friend
		_fixItTrait.fix_it(this->internal_mutex());  
	}

	template<typename mutexT>
	inline boost::uint32_t robust_shm_named_mutex<mutexT>::get_state()
	{  return this->internal_mutex().get_state();  }

}}  //xw::ipc

#endif   
