//////////////////////////////////////////////////////////////////////////////
//
// (C) Copyright Ion Gaztanaga 2005-2012. Distributed under the Boost
// Software License, Version 1.0. (See accompanying file
// LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//
// See http://www.boost.org/libs/interprocess for documentation.
//
//////////////////////////////////////////////////////////////////////////////
//
// Parts of the pthread code come from Boost Threads code:
//
//////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2001-2003
// William E. Kempf
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  William E. Kempf makes no representations
// about the suitability of this software for any purpose.
// It is provided "as is" without express or implied warranty.
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
//  (C) Copyright Ladislav Sopko, 3D Informatica s.r.l 2014
//  
//////////////////////////////////////////////////////////////////////////////

#if (defined _MSC_VER) && (_MSC_VER >= 1200)
#  pragma once
#endif

#ifndef XW_IPC_ROBUST_RECURSIVE_SPIN_MUTEX_HPP
#define XW_IPC_ROBUST_RECURSIVE_SPIN_MUTEX_HPP

#include <boost/interprocess/detail/config_begin.hpp>
#include <boost/interprocess/detail/workaround.hpp>
#include <boost/interprocess/detail/posix_time_types_wrk.hpp>
#include <boost/interprocess/detail/os_thread_functions.hpp>
#include <boost/interprocess/detail/atomic.hpp>
#include <boost/cstdint.hpp>
#include <boost/assert.hpp>
#include <xw-incl/ipc/robust_mutex.hpp>


#ifdef strRev
#	undef strRev
#endif
#define strRev "$Revision: 1.1 $"

namespace bipc = boost::interprocess;

namespace xw{ namespace ipc{

	class robust_recursive_mutex
	{
		typedef robust_recursive_mutex self_type;

		friend class fix_it_trait<self_type>;
		friend class set_name_trait<self_type>;

		robust_recursive_mutex(const robust_recursive_mutex &);
		robust_recursive_mutex &operator=(const robust_recursive_mutex &);
		
	public:

		robust_recursive_mutex();
		~robust_recursive_mutex();

		void lock();
		bool try_lock();
		bool timed_lock(const boost::posix_time::ptime &abs_time);
		void unlock();
		void take_ownership();
		const char* name() const;
		boost::uint32_t get_state(){
			return m_mutex.get_state();
		};	
		
	private:
		robust_mutex   m_mutex;
		unsigned int   m_nLockCount;
		volatile bipc::ipcdetail::OS_systemwide_thread_id_t   m_nOwner;		

		void fix_it(){
			m_mutex.fix_it();
		}
		void name(const char* n){
			m_mutex.name(n); 
		}
	};

	/************************************************************************/
	/*  IPC types                                                           */
	/************************************************************************/

	typedef robust_recursive_mutex robust_ipc_recursive_mutex;

	/************************************************************************/
	/*  IMPL                                                                */
	/************************************************************************/

	inline const char* robust_recursive_mutex::name() const { 
		return m_mutex._name; 
	}

	inline robust_recursive_mutex::robust_recursive_mutex()
		: m_nLockCount(0), m_nOwner(bipc::ipcdetail::get_invalid_systemwide_thread_id()){
		m_mutex._name[0] = 0;
	}

	inline robust_recursive_mutex::~robust_recursive_mutex(){}

	inline void robust_recursive_mutex::lock()
	{
		typedef bipc::ipcdetail::OS_systemwide_thread_id_t handle_t;
		const handle_t thr_id(bipc::ipcdetail::get_current_systemwide_thread_id());
		handle_t old_id;
		bipc::ipcdetail::systemwide_thread_id_copy(m_nOwner, old_id);
		if(bipc::ipcdetail::equal_systemwide_thread_id(thr_id , old_id)){
			if((unsigned int)(m_nLockCount+1) == 0){
				//Overflow, throw an exception
				XW_THROW_EXCEPTION(ex_ipc_generic, "Recursive lock overflow");
			}
			++m_nLockCount;
		}
		else{
			m_mutex.lock();
			bipc::ipcdetail::systemwide_thread_id_copy(thr_id, m_nOwner);
			m_nLockCount = 1;
		}
	}

	inline bool robust_recursive_mutex::try_lock()
	{
		typedef bipc::ipcdetail::OS_systemwide_thread_id_t handle_t;
		handle_t thr_id(bipc::ipcdetail::get_current_systemwide_thread_id());
		handle_t old_id;
		bipc::ipcdetail::systemwide_thread_id_copy(m_nOwner, old_id);
		if(bipc::ipcdetail::equal_systemwide_thread_id(thr_id , old_id)) {  // we own it
			if((unsigned int)(m_nLockCount+1) == 0){
				//Overflow, throw an exception
				XW_THROW_EXCEPTION(ex_ipc_generic, "Recursive lock overflow");
			}
			++m_nLockCount;
			return true;
		}
		if(m_mutex.try_lock()){
			bipc::ipcdetail::systemwide_thread_id_copy(thr_id, m_nOwner);
			m_nLockCount = 1;
			return true;
		}
		return false;
	}

	inline bool robust_recursive_mutex::timed_lock(const boost::posix_time::ptime &abs_time)
	{
		typedef bipc::ipcdetail::OS_systemwide_thread_id_t handle_t;
		if(abs_time == boost::posix_time::pos_infin){
			this->lock();
			return true;
		}
		const handle_t thr_id(bipc::ipcdetail::get_current_systemwide_thread_id());
		handle_t old_id;
		bipc::ipcdetail::systemwide_thread_id_copy(m_nOwner, old_id);
		if(bipc::ipcdetail::equal_systemwide_thread_id(thr_id , old_id)) {  // we own it
			if((unsigned int)(m_nLockCount+1) == 0){
				//Overflow, throw an exception
				XW_THROW_EXCEPTION(ex_ipc_generic, "Recursive lock overflow");
			}
			++m_nLockCount;
			return true;
		}
		if(m_mutex.timed_lock(abs_time)){
			bipc::ipcdetail::systemwide_thread_id_copy(thr_id, m_nOwner);
			m_nLockCount = 1;
			return true;
		}
		return false;
	}

	inline void robust_recursive_mutex::unlock()
	{
		typedef bipc::ipcdetail::OS_systemwide_thread_id_t handle_t;
		handle_t old_id;
		bipc::ipcdetail::systemwide_thread_id_copy(m_nOwner, old_id);
		const handle_t thr_id(bipc::ipcdetail::get_current_systemwide_thread_id());
		(void)old_id;
		(void)thr_id;
		BOOST_ASSERT(bipc::ipcdetail::equal_systemwide_thread_id(thr_id, old_id));
		--m_nLockCount;
		if(!m_nLockCount){
			const handle_t new_id(bipc::ipcdetail::get_invalid_systemwide_thread_id());
			bipc::ipcdetail::systemwide_thread_id_copy(new_id, m_nOwner);
			m_mutex.unlock();
		}
	}

	inline void robust_recursive_mutex::take_ownership()
	{
		typedef bipc::ipcdetail::OS_systemwide_thread_id_t handle_t;
		this->m_nLockCount = 1;
		const handle_t thr_id(bipc::ipcdetail::get_current_systemwide_thread_id());
		bipc::ipcdetail::systemwide_thread_id_copy(thr_id, m_nOwner);
	}


}}  //xw::ipc

#include <boost/interprocess/detail/config_end.hpp>

#endif   
